/**
 * Copyright 2011 J Adam Crain (jadamcrain@gmail.com)
 *
 * Licensed to J Adam Crain under one or more contributor license agreements.
 * See the NOTICE file distributed with this work for additional information
 * regarding copyright ownership. J Adam Crain licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License
 * at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.automatak.executor4s

object Result {

  def apply[A](fun: => A): Result[A] = {
    try { Success(fun) }
    catch { case ex: Exception => Failure(ex) }
  }

}

sealed trait Result[+A] {
  // throws any exceptions on the calling thread
  def get: A
  def isSuccess: Boolean
  def isFailure: Boolean
  def map[B](convert: A => B): Result[B]
  def flatMap[B](convert: A => Result[B]): Result[B]
  def foreach[B](fun: A => B): Unit
  def asOption: Option[Throwable]
}

final case class Success[A](value: A) extends Result[A] {
  override def get = value
  override def isSuccess = true
  override def isFailure = false
  override def map[B](convert: A => B) = Result(convert(value))
  override def flatMap[B](convert: A => Result[B]): Result[B] = {
    try {
      convert(value)
    } catch {
      case ex: Exception => Failure(ex)
    }
  }
  override def foreach[B](fun: A => B) = fun(value)
  override def asOption: Option[Throwable] = None
}

object Failure {
  def apply(msg: String): Failure = Failure(new Exception(msg))

  def exceptions(ts: Traversable[Throwable]): Failure = {
    Failure("Multiple failures detected: " + System.lineSeparator() + ts.mkString(System.lineSeparator()))
  }

  def failures(fs: Traversable[Failure]): Failure = {
    Failure("Multiple failures detected: " + System.lineSeparator() + fs.mkString(System.lineSeparator()))
  }
}

final case class Failure(ex: Throwable) extends Result[Nothing] {
  override def get = throw ex
  override def isSuccess = false
  override def isFailure = true
  override def map[B](convert: Nothing => B): Result[B] = this
  override def flatMap[B](convert: Nothing => Result[B]) = this
  override def foreach[B](fun: Nothing => B) = {}
  override def asOption: Option[Throwable] = Some(ex)
}