/**
 * Copyright 2011 J Adam Crain (jadamcrain@gmail.com)
 *
 * Licensed to J Adam Crain under one or more contributor license agreements.
 * See the NOTICE file distributed with this work for additional information
 * regarding copyright ownership. J Adam Crain licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License
 * at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.automatak.executor4s

import annotation.tailrec
import impl.DefaultAsynchronousStream
import java.util.concurrent.Executor
import collection.mutable.ListBuffer

trait Closeable {
  def close(): Unit
}

case object EndOfStream extends Exception

object AsynchronousStream {

  def open[A](implicit exe: Executor): AppendableAsynchronousStream[A] = new DefaultAsynchronousStream

  def success[A](a: A)(implicit exe: Executor): AsynchronousStream[A] = {
    val s = open[A]
    s.close(a)
    s
  }

  def success[A](as: Traversable[A])(implicit exe: Executor): AsynchronousStream[A] = {
    val s = open[A]
    as.foreach(s.append)
    s.close()
    s
  }

  def failure[A](t: Throwable)(implicit exe: Executor): AsynchronousStream[A] = {
    val s = open[A]
    s.close(t)
    s
  }

}

/**
 * Interface used to write to the backend of a stream
 * @tparam A
 */
trait Appendable[A] extends Closeable {

  /**
   * Appends an element to the stream
   * @return false if the stream is closed and cannot be appended, true otherwise
   * @param a
   */
  def append(a: A): Boolean

  /**
   * Close the stream with either an exception or final value
   * @param r
   */
  def close(r: Result[A]): Unit

  /**
   * Close the stream with it's final value
   * @param a
   */
  def close(a: A): Unit = close(Success(a))

  /**
   * Close the stream with an exception
   * @param ex
   */
  def close(ex: Throwable): Unit = close(Failure(ex))

  /**
   * Close the stream gracefully with EndOfStream
   */
  final override def close(): Unit = close(Failure(EndOfStream))

}

/**
 * Represents an asynchronous computation that produces a series of values of the same type
 * @tparam A
 */
trait AsynchronousStream[+A] extends Closeable {

  /**
   * @return Future[A] to the next element in the stream. Empty stream will return a failed future with EOS
   *
   */
  def next(): Future[A]

  /**
   * Converts a stream of one type into a stream of another type
   * @param f
   * @tparam B
   * @return
   */
  def map[B](f: A => B): AsynchronousStream[B] = {
    val that = this
    new AsynchronousStream[B] {
      def next(): Future[B] = that.next().map(f)
      def close(): Unit = that.close()
    }
  }

  /**
   * Asynchronously iterate over a stream, calling the same function
   */
  def listen[B](fun: Result[A] => B): Unit = {
    def recurse(): Unit = next().listen { res =>
      res match {
        case s: Success[A] =>
          fun(s)
          recurse()
        case f: Failure =>
          fun(f)
      }
    }
    recurse()
  }

  /**
   * Synchronously iterate over the stream. Blocking function used primarily for testing.
   * @param fun
   * @tparam B
   */
  def foreach[B](fun: Result[A] => B): Unit = {
    @tailrec
    def recurse(): Unit = {
      next().await match {
        case s: Success[A] =>
          fun(s)
          recurse()
        case f: Failure =>
          fun(f)
      }
    }
    recurse()
  }

  /**
   * Synchronously create a list of all elements in the stream. Blocking function used primarily for testing.
   */
  def toList: List[Result[A]] = {
    val builder = ListBuffer.empty[Result[A]]
    foreach(x => builder.append(x))
    builder.toList
  }

}

trait AppendableAsynchronousStream[A] extends AsynchronousStream[A] with Appendable[A]