/**
 * Copyright 2011 J Adam Crain (jadamcrain@gmail.com)
 *
 * Licensed to J Adam Crain under one or more contributor license agreements.
 * See the NOTICE file distributed with this work for additional information
 * regarding copyright ownership. J Adam Crain licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License
 * at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.automatak.executor4s.impl

import com.automatak.executor4s._
import java.util.concurrent.Executor
import annotation.tailrec

private[executor4s] final class DefaultAsynchronousStream[A](implicit exe: Executor) extends AppendableAsynchronousStream[A] {

  private val mutex = new Object
  private val values = collection.mutable.Queue.empty[A]
  private val futures = collection.mutable.Queue.empty[Settable[A]]
  private var closed = Option.empty[Throwable]

  override def next(): Future[A] = {
    val f = Future.pending[A]
    mutex.synchronized {
      if (values.isEmpty) {
        closed match {
          case Some(x) => f.setFailure(x)
          case None => futures.enqueue(f)
        }
      } else {
        f.setSuccess(values.dequeue())
      }
    }
    f
  }

  override def append(a: A): Boolean = mutex.synchronized(appendPriv(a))

  override def close(r: Result[A]): Unit = mutex.synchronized {
    r match {
      case Success(a) =>
        appendPriv(a)
        closePriv(EndOfStream)
      case Failure(ex) =>
        closePriv(ex)
    }
  }

  private def appendPriv(a: A): Boolean = {
    if (closed.isDefined) false
    else {
      if (futures.isEmpty) values.enqueue(a)
      else futures.dequeue().setSuccess(a)
      true
    }
  }

  private def closePriv(ex: Throwable): Unit = {
    if (closed.isEmpty) {
      closed = Some(ex)
      futures.dequeueAll(_ => true).foreach(s => dispatch(s.setFailure(ex)))
    }
  }

}
