/**
 * Copyright 2011 J Adam Crain (jadamcrain@gmail.com)
 *
 * Licensed to J Adam Crain under one or more contributor license agreements.
 * See the NOTICE file distributed with this work for additional information
 * regarding copyright ownership. J Adam Crain licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License
 * at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.automatak.executor4s.impl

import java.lang.IllegalStateException

import com.automatak.executor4s._
import java.util.concurrent.Executor
import annotation.tailrec

final class DefaultFuture[A](private var value: Option[Result[A]] = None)(implicit dispatcher: Executor) extends SettableFuture[A] {

  private val mutex = new Object
  private val listeners = collection.mutable.Set.empty[Result[A] => Unit]
  private val onSetListeners = collection.mutable.Set.empty[Result[A] => Unit]

  override def replicate[B] = new DefaultFuture[B]()
  override def replicate[B](b: B) = new DefaultFuture[B](Some(Success(b)))

  override def await(): Result[A] = mutex.synchronized {
    @tailrec
    def get(): Result[A] = value match {
      case Some(x) => x
      case None =>
        mutex.wait()
        get()
    }
    get()
  }

  override def listen(fun: Result[A] => Unit): Unit = mutex.synchronized {
    value match {
      case Some(x) => dispatch(fun(x))
      case None => listeners.add(fun)
    }
  }

  override def setSuccess(a: A) = set(Success(a))

  override def setFailure(t: Throwable) = set(Failure(t))

  override def set(result: Result[A]) = {
    mutex.synchronized {
      value match {
        case Some(x) => throw new IllegalStateException("Future has already been set to: " + value)
        case None =>
          value = Some(result)
          mutex.notifyAll()
      }

      onSetListeners.foreach(fun => fun(result))
      onSetListeners.clear()
      listeners.foreach(fun => dispatch(fun(result)))
      listeners.clear()
    }
  }

  override def onSet(fun: Result[A] => Unit): Unit = mutex.synchronized {
    value match {
      case Some(x) => fun(x)
      case None => onSetListeners.add(fun)
    }
  }

}